//////////////////////////////////////////////////////////////////////////////////////////////////////
//
// D3D8 Bitmap Font Class, 1.01
// Greg Wade, September 30th, 2001
// greg@wadesoft.com
//
//////////////////////////////////////////////////////////////////////////////////////////////////////
//
// DISTRIBUTION
//	
//	There isn't a copywrite; just be sensible and if you distribute this code, please indicate your
//	changes in the code. If you feel there are significant enhancements to be made, please let me
//	know of them. This code is in the public domain by my own decree.
//
// PURPOSE
//
//	The purpose of this class is to efficiently draw 2D pre-generated fonts stored in a texture on 
//	the screen in while in 3D mode (there is no other choice with D3D8), for fps display, etc.
//
// REQUIRED
//
//	-	LMNOPC.COM's Bitmap Font Builder to create the texture file and the character width file.
//		(you could do this by hand instead, by providing the following files manually).
//	-	A 256x256, 512x512, etc, texture file WITH an alpha channel (created by BFB above).
//	-	A 256 byte file containing the width in pixels of each character (created by BFB above).
//
// FEATURES
//
//	-	Efficient drawing method compared to d3dx's font class.
//	-	Cross platform concept (fonts done in this way can be ported to OpenGL, etc).
//	-	Batching of text drawing using begin/end method calls for efficiency.
//	-	Friendly application regarding the rendering state - restores the previous rendering state.
//	-	Clean method-call based configuration similar to that used in d3d8 and opengl.
//	-	Loads almost any type of texture file (that supports an alpha channel) including dds,tga,png.
//	-	Automatically converts the texture into an optimal color depth for the primary adapter.
//	-	Converts the texture into a d3dx compressed texture if hardware supports it. 32k for 256x256!
//	-	Optionally performs vertex alpha blending on the text (unrelated to the texture alpha chan)
//	-	Optionally blends textures from one color at the top to another at the bottom. Alpha too!
//	-	Can display an unlimited number of characters with any size of vertex buffer (refills it).
//	-	Uses an efficient 4 vertices per character using triangle strips.
//
// USAGE
//
//	1.	Use LMNOPC.COM's Bitmap Font Builder to create a left justified font texture. Save the file
//		as a 32 bit TGA texture because it contains the alpha channel we use. Also, make sure you
//		select the option to generate the full ASCII character set.
//	2.	Use BFB (above) to create a binary character width data file to go with the texture. 
//	3.	In your code, construct an instance of this class using the NEW operator.
//	4.	Use the Load method to load the texture (passing the BL_BITMAP flag).
//	5.	Use the Load method to load the character widths (using the BL_SIZES flag).
//	6.	Now you can render your text with the DrawText or DrawTextEx methods.
//	7.	If you make more than one call to the DrawText method(s), it is best to surround your calls
//		with the BeginDraw and EndDraw method calls for the sake of efficiency.
//	8.	When you are all done, destroy the object using the delete operator.
//
// METHODS & PROPERTIES
//
//	int DrawText(const char* text, int x, int y, DWORD color);
//
//		- Draws text with x, y, being the top left hand corner of the text rectangle.
//		- Uses color as the solid drawing color. 0xaarrggbb
//		- No real return
//		
//	int DrawTextEx(const char* text, int x, int y, DWORD colorTop, DWORD colorBottom, float scale);
//
//		- Draws text with x, y being the top left hand corner of the text rectangle.
//		- Uses colorTop as the top color for blending, and colorBottom for the bottom.
//		- Alpha is blended along with color values
//		- Scale is used as a multiplier for the normal font size.
//		- No real return 
//
//	int Load(BitmapFontLoadEnum filetype, char* fn);
//
//		- BL_BITMAP as the filetype informs the class to load a texture (tga, png, dds, etc)
//		- BL_SIZES as the filetype loads the 256 bytes binary font width file
//		- This method should be called for both types of files prior to using the draw methods.
//		- Returns > 0 on success
//
//	int BeginDraw();
//
//		- Helps batch drawing performance. Call this before drawing all your text. 
//		- Not required, but suggested (called for you automatically if you didn't, but it's slower)
//		- No real return.
//
//	int EndDraw();
//
//		- Companion to BeginDraw. 
//		- Called after all your text drawing is done.
//
//	RECT GetTextRectEx(const char* text, int x, int y, float scale);
//
//		- Get the rectangle around the text, considering scale (used with DrawTextEx).
//		- Good to know if text is visible.
//
//	RECT GetTextRect(const char* text, int x, int y);
//	
//		- Get the rectangle around the text (used with DrawText)
//		- Good to know if text is visible.
//
// FUTURE ENHANCEMENTS
//
//	-	More robust error checking.
//	-	Automatic word wrapping.
//	-	Text scrolling.
//	-	Text file reading with special tags for headers, etc.
//
//////////////////////////////////////////////////////////////////////////////////////////////////////


// our vertex buffer definition
//#define D3DFVF_FONT (D3DFVF_XYZRHW|D3DFVF_DIFFUSE|D3DFVF_TEX1)
/*
struct FONT_VERTEX
{
    FLOAT x, y, z, rhw; 
    DWORD color;        
	float tu, tv;
};
*/

// an enumeration used for the load method. what type of file are we loading?
enum BitmapFontLoadEnum
{
	BL_BITMAP,
	BL_SIZES
};


// vertex buffer size - each character requires 4 bytes
const FONTVBSIZE = 4*1024;


// our font class
class CBitmapFont
{
protected:
	LPDIRECT3D8 d3d;
	LPDIRECT3DDEVICE8 d3dDevice;
	LPDIRECT3DTEXTURE8 tex;
	LPDIRECT3DVERTEXBUFFER8 vb;
	FONT_VERTEX *v;
	
	int	bmWidth, bmHeight;
	int cellWidth, cellHeight;
	char charWidths[256];
	DWORD oldState;

	long numChars;
	bool begun;

public:
	CBitmapFont(LPDIRECT3D8 pd3d, LPDIRECT3DDEVICE8 pd3dDevice);
	~CBitmapFont();

	int DrawText(const char* text, int x, int y, DWORD color);
	int DrawTextEx(const char* text, int x, int y, DWORD colorTop, DWORD colorBottom, float scale);
	int Load(BitmapFontLoadEnum filetype, char* fn);
	int BeginDraw();
	int EndDraw();
	RECT GetTextRectEx(const char* text, int x, int y, float scale);
	RECT GetTextRect(const char* text, int x, int y);

private:
	D3DFORMAT GetBestFormat();
};


// constructor
inline CBitmapFont::CBitmapFont(LPDIRECT3D8 pd3d, LPDIRECT3DDEVICE8 pd3dDevice)
{
	// remember our d3d object
	d3dDevice= pd3dDevice;
	d3d= pd3d;

	// initial settings so we know what we have and have not loaded
	tex= 0;
	bmHeight= 0;
	bmWidth= 0;
	cellHeight= 0;
	cellWidth= 0;
	oldState= 0;
	vb= 0;
	begun= false;
	numChars= 0;

	// fill our character width buffer with default values
	strnset(charWidths, 16, 256);

	// create the vertex buffer
	d3dDevice->CreateVertexBuffer( FONTVBSIZE*sizeof(FONT_VERTEX), 0, D3DFVF_FONT, D3DPOOL_SYSTEMMEM, &vb);

	// the rendering state block is created here
	d3dDevice->CreateStateBlock(D3DSBT_ALL, &oldState);
}


// destructor
inline CBitmapFont::~CBitmapFont()
{
	if (vb) vb->Release();
	if (tex) tex->Release();
	if (oldState) d3dDevice->DeleteStateBlock(oldState);
}


// get a rectangle for the text, considering a possible text scaling.
// useful to know if some text is visible or not
inline RECT CBitmapFont::GetTextRectEx(const char* text, int x, int y, float scale)
{
	RECT r;
	
	// start the top left corner at the ones passed in
	r.left= x;
	r.top= y;

	// add the height right away
	r.bottom= y+ long((float)(cellHeight)*float(scale));

	// start the right at the same as the left, but increase for each character
	r.right= x;
	for (unsigned long t=0; t<strlen(text); t++)
	{
		r.right+= long(float(charWidths[text[t]]) * scale);
	}

	return r;
}


// get a rectangle for the text. see GetTextRectEx
inline RECT CBitmapFont::GetTextRect(const char* text, int x, int y)
{
	return GetTextRectEx(text, x, y, 1.0f);
}


// begin the batch drawing process. called automatically for you if you forgot, but this does not leverage the
// efficiency of batching text unless you call it yourself along with EndDraw afterwards.
inline int CBitmapFont::BeginDraw()
{
	// initialize some state variables
	numChars= 0;
	begun= true;

	// remember the rendering state so we can put it back
	d3dDevice->CaptureStateBlock(oldState);

	// set our critical blending params
    d3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, true );
    d3dDevice->SetRenderState( D3DRS_SRCBLEND,   D3DBLEND_SRCALPHA );
    d3dDevice->SetRenderState( D3DRS_DESTBLEND,  D3DBLEND_INVSRCALPHA );
    d3dDevice->SetTextureStageState( 0, D3DTSS_COLOROP,   D3DTOP_MODULATE );
    d3dDevice->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
    d3dDevice->SetTextureStageState( 0, D3DTSS_COLORARG2, D3DTA_DIFFUSE );
    d3dDevice->SetTextureStageState( 0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE );
    d3dDevice->SetTextureStageState( 0, D3DTSS_ALPHAARG2, D3DTA_DIFFUSE );
    d3dDevice->SetTextureStageState( 0, D3DTSS_ALPHAOP,   D3DTOP_MODULATE );

	// use point filtering since we want sharp text
	d3dDevice->SetTextureStageState( 0, D3DTSS_MINFILTER, D3DTEXF_POINT );
    d3dDevice->SetTextureStageState( 0, D3DTSS_MAGFILTER, D3DTEXF_POINT );

	// set our other required generic params
    d3dDevice->SetRenderState( D3DRS_ALPHAREF,         0x08 );
    d3dDevice->SetRenderState( D3DRS_ALPHAFUNC,   D3DCMP_GREATEREQUAL);
    d3dDevice->SetRenderState( D3DRS_FILLMODE,   D3DFILL_SOLID );
    d3dDevice->SetRenderState( D3DRS_CULLMODE,   D3DCULL_CCW );
    d3dDevice->SetRenderState( D3DRS_ZENABLE,          FALSE );
    d3dDevice->SetRenderState( D3DRS_STENCILENABLE,    FALSE );
    d3dDevice->SetRenderState( D3DRS_CLIPPING,         TRUE );
    d3dDevice->SetRenderState( D3DRS_EDGEANTIALIAS,    FALSE );
    d3dDevice->SetRenderState( D3DRS_CLIPPLANEENABLE,  FALSE );
    d3dDevice->SetRenderState( D3DRS_VERTEXBLEND,      FALSE );
    d3dDevice->SetRenderState( D3DRS_INDEXEDVERTEXBLENDENABLE, FALSE );
    d3dDevice->SetRenderState( D3DRS_FOGENABLE,        FALSE );
	d3dDevice->SetTextureStageState( 0, D3DTSS_MIPFILTER, D3DTEXF_NONE );
    d3dDevice->SetTextureStageState( 0, D3DTSS_TEXCOORDINDEX, 0 );
    d3dDevice->SetTextureStageState( 0, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_DISABLE );
    d3dDevice->SetTextureStageState( 1, D3DTSS_COLOROP,   D3DTOP_DISABLE );
    d3dDevice->SetTextureStageState( 1, D3DTSS_ALPHAOP,   D3DTOP_DISABLE );

	// apply the texture
	d3dDevice->SetTexture(0, tex);

	// prepare to draw the text
    d3dDevice->SetStreamSource( 0, vb, sizeof(FONT_VERTEX) );
    d3dDevice->SetVertexShader( D3DFVF_FONT );

	// lock the vertex buffer
	vb->Lock( 0, 0, (BYTE**)&v, 0 );  
	return 1;
}


// companion of the BeginDraw function. Surround your text drawing with these function calls
inline int CBitmapFont::EndDraw()
{
	// set some state variables
	begun= false;

	// unlock the vertices
	vb->Unlock();

	// draw the actual characters now. each character is 3 strips long, forming a quad of 4 vertices.
	for (long cpos=0; cpos<numChars; cpos++)
	{
		d3dDevice->DrawPrimitive( D3DPT_TRIANGLESTRIP, cpos*4, 3 );
	}

	// put back the old rendering state variables
	d3dDevice->ApplyStateBlock(oldState);
	return 1;
}


// function used to check if your hardware supports a requested texture format while in another
inline BOOL IsTextureFormatOk( LPDIRECT3D8 d3d, D3DFORMAT TextureFormat, D3DFORMAT AdapterFormat ) 
{
	HRESULT hr = d3d->CheckDeviceFormat( D3DADAPTER_DEFAULT,
                                          D3DDEVTYPE_HAL,
                                          AdapterFormat,
                                          0,
                                          D3DRTYPE_TEXTURE,
                                          TextureFormat);

    return SUCCEEDED( hr );
}


// function to intelligently find the best texture format for the font texture. attempts to find
// the least memory hungry format available. almost all cards support 8x8x8x8 32 bit, and this is
// the last resort due to huge memory use.
inline D3DFORMAT CBitmapFont::GetBestFormat()
{
	D3DDISPLAYMODE display;
	D3DFORMAT format;

	// get the adapter display mode
	d3d->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &display); 

	// start with the standard 8x8x8x8 format. we will check for better options next
	format= D3DFMT_A8R8G8B8;

	// next try a 16 bit format with alpha
	if (IsTextureFormatOk(d3d, D3DFMT_A1R5G5B5, display.Format)) format= D3DFMT_A1R5G5B5;

	// next try another 16 bit format with alpha
	if (IsTextureFormatOk(d3d, D3DFMT_A4R4G4B4, display.Format)) format= D3DFMT_A4R4G4B4;

	// try for the ultimate - compressed texture
	if (IsTextureFormatOk(d3d, D3DFMT_DXT1, display.Format)) format= D3DFMT_DXT1;

	// return the result
	return format;
}


// returns the font height in the case of a bitmap load. note does not use the real bitmap height but rather
// the height of the bitmap / 16. generally returns > 0 for success.
inline int CBitmapFont::Load(BitmapFontLoadEnum filetype, char* fn)
{
	int res= 1;
	D3DXIMAGE_INFO info;

	switch(filetype)
	{
		case BL_BITMAP:
			if (!FAILED(D3DXCreateTextureFromFileEx( d3dDevice, fn, NULL, NULL, 1, 0, GetBestFormat(), D3DPOOL_MANAGED,
				D3DX_DEFAULT, D3DX_DEFAULT, 0, &info, NULL, &tex )))
			{
				D3DSURFACE_DESC s;
				tex->GetLevelDesc(0, &s);
				bmWidth= s.Width;
				bmHeight= s.Height;
				cellHeight= bmHeight/16;
				cellWidth= bmWidth/16;
				res= cellHeight;
			} 
			else res= 0;

			break;

		case BL_SIZES:
			FILE *fil;
			fil= fopen(fn, "rb");
			if (fread(charWidths, 1, 256, fil) != 256) res= 0;		
			fclose(fil);
			break;
	}
	
	return res;
}


// draw text at x, y using color
inline int CBitmapFont::DrawText(const char* text, int x, int y, DWORD color)
{
	return DrawTextEx(text, x, y, color, color, 1.0f);
}


// draw text at x, y using colorTop blended down to colorBottom, with size multiplied by scale
inline int CBitmapFont::DrawTextEx(const char* text, int x, int y, DWORD colorTop, DWORD colorBottom, float scale)
{
	unsigned int chr=0;
	unsigned int cpos= 0;
	float screenX1= (float)x;
	float screenY1= (float)y;
	float screenX2= 0.0f;
	float screenY2= 0.0f;
	bool wasBegun= true;

	// abort now if the bitmap was not loaded or if the string is empty
	if ((tex==0) || (strlen(text)==0)) return 0;

	// we should abort if our vertex buffer simply isn't big enough to handle this many characters
	if (strlen(text)*4 > FONTVBSIZE) return 0;	

	// if our buffer is big enough, but doesn't have enough room at the moment, we should purge it.
	// if that is the case it also means that begin/end is being handled by the user since numChars > 0 in this case
	if ((((numChars+strlen(text))*4) > FONTVBSIZE) && (numChars > 0))
	{
		EndDraw();
		BeginDraw();
	}

	// if it was not already done, then we should call begin. it should have been called outside to batch
	// the drawing of characters for efficiency, but if not, we will do it here. we must remember this action and
	// then automatically call end
	if (begun==false)
	{
		BeginDraw();
		wasBegun= false;
	}

	// cycle through all the characters in the string
	for (cpos=numChars; cpos<strlen(text)+numChars; cpos++)
	{
		// isolate the character we are interested in
		chr= text[cpos-numChars];

		// screen space dimensions for this character
		screenX2= screenX1+ (float)charWidths[chr] * scale;
		screenY2= screenY1+ (float)cellHeight * scale;

		// which part of the bitmap do we want for this character, scaled from 0 to 1
		float bmy1= float( (cellHeight)*(chr/16) ) / float(bmHeight);
		float bmx1= float( (cellWidth)*(chr%16) ) / float(bmWidth); 
		float bmy2= bmy1 + (float(cellHeight)) / (float(bmHeight));
		float bmx2= bmx1 + (float(charWidths[chr])) / (float(bmWidth));

		// set the vertex buffer data for each character.
		// each character is actually a triangle strip with 4 vertices (3 strips) making a square
		v[cpos*4+0].x= screenX1-0.5f;	v[cpos*4+0].y= screenY2-0.5f;	v[cpos*4+0].z= 0.9f; v[cpos*4+0].rhw= 1.0f; v[cpos*4+0].color= colorBottom; v[cpos*4+0].tu= bmx1; v[cpos*4+0].tv= bmy2;
		v[cpos*4+1].x= screenX1-0.5f;	v[cpos*4+1].y= screenY1-0.5f;	v[cpos*4+1].z= 0.9f; v[cpos*4+1].rhw= 1.0f; v[cpos*4+1].color= colorTop;	v[cpos*4+1].tu= bmx1; v[cpos*4+1].tv= bmy1;
		v[cpos*4+2].x= screenX2-0.5f;	v[cpos*4+2].y= screenY2-0.5f;	v[cpos*4+2].z= 0.9f; v[cpos*4+2].rhw= 1.0f; v[cpos*4+2].color= colorBottom; v[cpos*4+2].tu= bmx2; v[cpos*4+2].tv= bmy2;
		v[cpos*4+3].x= screenX2-0.5f;	v[cpos*4+3].y= screenY1-0.5f;	v[cpos*4+3].z= 0.9f; v[cpos*4+3].rhw= 1.0f; v[cpos*4+3].color= colorTop;	v[cpos*4+3].tu= bmx2; v[cpos*4+3].tv= bmy1;
		
		if(chr == '\n')
		{
			screenX1 = (float)x;
			screenY1 += cellHeight;
		}
		else
		{
			// increase the starting position of the next character
			screenX1+= charWidths[chr] * scale;
		}
	}

	// increase the number of characters that will be drawn
	numChars+= strlen(text);

	// if begin was not called, then the user must be using this class inefficiently without batching, so we should
	// call end now
	if (wasBegun==false)
	{
		EndDraw();
	}
	return 1;
}