#include "StdAfx.h"

#define		DISPLAY_API	__declspec(dllexport)

#include "../Common/Common.h"
#include "interfaces.h"
#include "manager.h"

#include ".\x3dfont.h"

// constructor
 X3DFont::X3DFont(LPDIRECT3D8 pd3d, LPDIRECT3DDEVICE8 pd3dDevice)
{
	// remember our d3d object
	m_d3dDevice= pd3dDevice;
	d3d= pd3d;

	// initial settings so we know what we have and have not loaded
	m_texFont = NULL;
	m_pVB = NULL;
	bmHeight= 0;
	bmWidth= 0;
	cellHeight= 0;
	cellWidth= 0;	
	begun= false;
	numChars= 0;

	// fill our character width buffer with default values
	strnset(charWidths, 16, 256);

	// create the vertex buffer
	if(S_OK == CreateDisplayObject(DISIID_IVertexBuffer, (void**) &m_pVB))
	{
		m_pVB->Create(FONTVBSIZE*sizeof(FONT_VERTEX), D3DUSAGE_WRITEONLY, D3DFVF_FONT, D3DPOOL_DEFAULT);
	}	
	// create the vertex buffer
	if(S_OK == CreateDisplayObject(DISIID_IIndexBuffer, (void**) &m_pIB))
	{
		m_pIB->Create(FONTIBSIZE*sizeof(WORD), D3DUSAGE_WRITEONLY, D3DFMT_INDEX16, D3DPOOL_DEFAULT);
	}	
		
	m_dwRef = 0;
}


// destructor
 X3DFont::~X3DFont()
{
	SAFE_RELEASE(m_texFont);
	SAFE_RELEASE(m_pVB);
	SAFE_RELEASE(m_pIB);
}


HRESULT		X3DFont::AddRef()
{
	return m_dwRef++;
}

HRESULT		X3DFont::Release()
{

	if(m_dwRef > 0)
	{
		m_dwRef--;
	}
	if(m_dwRef == 0)
	{	
		delete this;
		TRACE("X3DFont object destroyed.");
		return 0;
	}	
	else
	{
		return m_dwRef;
	}	
}

// returns the font height in the case of a bitmap load. note does not use the real bitmap height but rather
// the height of the bitmap / 16. generally returns > 0 for success.
 HRESULT X3DFont::Load(BitmapFontLoadEnum filetype, char* fn)
{
	FILE *sizes = NULL;
	DWORD dwRet = S_FALSE;
	WORD *pwInd; int base = 0;
	char cFont[MAX_PATH];
	switch(filetype)
	{
		case BL_BITMAP:
			if(S_OK == CreateDisplayObject(DISIID_ITexture,(void**) &m_texFont))
			{
				dwRet = m_texFont->LoadTextureFromFileEx(fn, 1, GetBestFormat(), D3DPOOL_MANAGED);
				if(dwRet != S_OK)
				{
					XException exp("Cannot load texture for font.", dwRet);
					THROW(exp);
				}
				bmWidth = m_texFont->Width();
				bmHeight = m_texFont->Height();
				cellHeight = bmHeight / 16;
				cellWidth = bmWidth / 16;									
				m_pIB->GetBuffer()->Lock(0,0, (BYTE**) &pwInd, 0);
				for(int i = 0; i < FONTIBSIZE; i += 6)
				{
					pwInd[i + 0] = 0 + base;
					pwInd[i + 1] = 1 + base;
					pwInd[i + 2] = 2 + base;

					pwInd[i + 3] = 1 + base;
					pwInd[i + 4] = 3 + base;
					pwInd[i + 5] = 2 + base;

					base += 4;
				}
				m_pIB->GetBuffer()->Unlock();				
			}
			break;

		case BL_SIZES:			
			cmnGetDataFilePath(cFont, fn);
			sizes = fopen(cFont, "rb");
			if(fread(charWidths, 1, 256, sizes) != 256)
			{
				XException exp("D3DXFont: Sizes file for font is invalid.");
				THROW(exp);						
			}
			fclose(sizes);
			dwRet = S_OK;
			break;
	}	
	return dwRet;
}

// begin the batch drawing process. called automatically for you if you forgot, but this does not leverage the
// efficiency of batching text unless you call it yourself along with EndDraw afterwards.
 HRESULT X3DFont::BeginDraw()
{
	// initialize some state variables
	numChars= 0;
	begun= true;
	
	// set our critical blending params
	m_d3dDevice->SetRenderState(D3DRS_LIGHTING, FALSE);
    m_d3dDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
    m_d3dDevice->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
    m_d3dDevice->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA);   

	// apply the texture
	m_d3dDevice->SetTexture(0, m_texFont->GetTexture());

	// prepare to draw the text
	m_d3dDevice->SetStreamSource(0, m_pVB->GetBuffer(), sizeof(FONT_VERTEX));
	m_d3dDevice->SetIndices(m_pIB->GetBuffer(), 0);
    m_d3dDevice->SetVertexShader(D3DFVF_FONT);

	// lock the vertex buffer
	m_pVB->GetBuffer()->Lock(0, 0, (BYTE**)&v, 0);  
	return 1;
}

// companion of the BeginDraw function. Surround your text drawing with these function calls
 HRESULT X3DFont::EndDraw()
{
	// set some state variables
	begun= false;

	// unlock the vertices
	m_pVB->GetBuffer()->Unlock();
	
	// draw the actual characters now. each character is 3 strips long, forming a quad of 4 vertices.	
	m_d3dDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, numChars*4, 0, numChars*2);
		
	m_d3dDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
	m_d3dDevice->SetRenderState(D3DRS_LIGHTING, TRUE);

	return 1;
}

// draw text at x, y using color
 HRESULT X3DFont::DrawText(const char* text, int x, int y, DWORD color)
{
	return DrawTextEx(text, x, y, color, color, 1.0f);
}


// draw text at x, y using colorTop blended down to colorBottom, with size multiplied by scale
 HRESULT X3DFont::DrawTextEx(const char* text, int x, int y, DWORD colorTop, DWORD colorBottom, float scale)
{
	int chr=0;
	int cpos= 0;
	int len = strlen(text);
	float screenX1= (float)x;
	float screenY1= (float)y;
	float screenX2= 0.0f;
	float screenY2= 0.0f;
	bool wasBegun= true;

	
	// abort now if the bitmap was not loaded or if the string is empty
	// we should abort if our vertex buffer simply isn't big enough to handle this many characters
	if ((len == 0) || (len > MAX_STRING_LEN))
	{
		return 0;	
	}

	// if our buffer is big enough, but doesn't have enough room at the moment, we should purge it.
	// if that is the case it also means that begin/end is being handled by the user since numChars > 0 in this case
	if ((((numChars+len)*4) > FONTVBSIZE) && (numChars > 0))
	{
		EndDraw();
		BeginDraw();
	}

	// if it was not already done, then we should call begin. it should have been called outside to batch
	// the drawing of characters for efficiency, but if not, we will do it here. we must remember this action and
	// then automatically call end
	if (begun==false)
	{
		BeginDraw();
		wasBegun= false;
	}
	
	// cycle through all the characters in the string
	for (cpos=numChars; cpos < len+numChars; cpos++)
	{
		// isolate the character we are interested in
		chr= text[cpos-numChars];

		// screen space dimensions for this character
		screenX2= screenX1+ (float)charWidths[chr] * scale;
		screenY2= screenY1+ (float)cellHeight * scale;

		// which part of the bitmap do we want for this character, scaled from 0 to 1
		float bmy1= float( (cellHeight)*(chr/16) ) / float(bmHeight);
		float bmx1= float( (cellWidth)*(chr%16) ) / float(bmWidth); 
		float bmy2= bmy1 + (float(cellHeight)) / (float(bmHeight));
		float bmx2= bmx1 + (float(charWidths[chr])) / (float(bmWidth));

		// set the vertex buffer data for each character.
		// each character is actually a triangle strip with 4 vertices (3 strips) making a square
		v[cpos*4+0].x= screenX1-0.5f;	v[cpos*4+0].y= screenY2-0.5f;	v[cpos*4+0].z= 0.0f; v[cpos*4+0].rhw= 1.0f; v[cpos*4+0].color= colorBottom; v[cpos*4+0].tu= bmx1; v[cpos*4+0].tv= bmy2;
		v[cpos*4+1].x= screenX1-0.5f;	v[cpos*4+1].y= screenY1-0.5f;	v[cpos*4+1].z= 0.0f; v[cpos*4+1].rhw= 1.0f; v[cpos*4+1].color= colorTop;	v[cpos*4+1].tu= bmx1; v[cpos*4+1].tv= bmy1;
		v[cpos*4+2].x= screenX2-0.5f;	v[cpos*4+2].y= screenY2-0.5f;	v[cpos*4+2].z= 0.0f; v[cpos*4+2].rhw= 1.0f; v[cpos*4+2].color= colorBottom; v[cpos*4+2].tu= bmx2; v[cpos*4+2].tv= bmy2;
		v[cpos*4+3].x= screenX2-0.5f;	v[cpos*4+3].y= screenY1-0.5f;	v[cpos*4+3].z= 0.0f; v[cpos*4+3].rhw= 1.0f; v[cpos*4+3].color= colorTop;	v[cpos*4+3].tu= bmx2; v[cpos*4+3].tv= bmy1;
		
		if(chr == '\n')
		{
			screenX1 = (float)x;
			screenY1 += cellHeight;
		}
		else
		{
			// increase the starting position of the next character
			screenX1+= charWidths[chr] * scale;
		}
	}
	
	// increase the number of characters that will be drawn	
	numChars += len;

	// if begin was not called, then the user must be using this class inefficiently without batching, so we should
	// call end now
	if (wasBegun==false)
	{
		EndDraw();
	}
	return 1;
}

// get a rectangle for the text, considering a possible text scaling.
// useful to know if some text is visible or not
 RECT X3DFont::GetTextRectEx(const char* text, int x, int y, float scale)
{
	RECT r;
	
	// start the top left corner at the ones passed in
	r.left= x;
	r.top= y;

	// add the height right away
	r.bottom= y+ long((float)(cellHeight)*float(scale));

	// start the right at the same as the left, but increase for each character
	r.right= x;
	for (unsigned long t=0; t<strlen(text); t++)
	{
		r.right+= long(float(charWidths[text[t]]) * scale);
	}

	return r;
}


// get a rectangle for the text. see GetTextRectEx
 RECT X3DFont::GetTextRect(const char* text, int x, int y)
{
	return GetTextRectEx(text, x, y, 1.0f);
}


// function used to check if your hardware supports a requested texture format while in another
 BOOL IsTextureFormatOk( LPDIRECT3D8 d3d, D3DFORMAT TextureFormat, D3DFORMAT AdapterFormat ) 
{
	HRESULT hr = d3d->CheckDeviceFormat( D3DADAPTER_DEFAULT,
                                          D3DDEVTYPE_HAL,
                                          AdapterFormat,
                                          0,
                                          D3DRTYPE_TEXTURE,
                                          TextureFormat);

    return SUCCEEDED( hr );
}


// function to intelligently find the best texture format for the font texture. attempts to find
// the least memory hungry format available. almost all cards support 8x8x8x8 32 bit, and this is
// the last resort due to huge memory use.
 D3DFORMAT X3DFont::GetBestFormat()
{
	D3DDISPLAYMODE display;
	D3DFORMAT format;

	// get the adapter display mode
	d3d->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &display); 

	// start with the standard 8x8x8x8 format. we will check for better options next
	format= D3DFMT_A8R8G8B8;

	// next try a 16 bit format with alpha
	if (IsTextureFormatOk(d3d, D3DFMT_A1R5G5B5, display.Format)) format= D3DFMT_A1R5G5B5;

	// next try another 16 bit format with alpha
	if (IsTextureFormatOk(d3d, D3DFMT_A4R4G4B4, display.Format)) format= D3DFMT_A4R4G4B4;

	// try for the ultimate - compressed texture
	if (IsTextureFormatOk(d3d, D3DFMT_DXT1, display.Format)) format= D3DFMT_DXT1;

	// return the result
	return format;
}