#include "StdAfx.h"

#define		DISPLAY_API	__declspec(dllexport)


#include "../Common/Common.h"
#include "interfaces.h"
#include "defs.h"
#include "manager.h"
#include "xtexture.h"
#include "xmesh.h"
#include "xvertexbuffer.h"
#include "xindexbuffer.h"
#include "xresourcemanager.h"


XMesh::XMesh(void)
{
	m_lpMesh = NULL;
	m_lpDevice = NULL;
	m_bVisible = FALSE;		
	D3DXMatrixIdentity(&m_TransMat);
	m_dwRef = 0;
	m_dwNumMat = 0;
	m_pMaterials = NULL;
	m_ppTextures = NULL;
	m_dwID = -1;
}

XMesh::~XMesh(void)
{
	if(m_lpDevice)
	{
		ReleaseResource();
		if(m_ppTextures)
		{
			for(int i = 0; i < (int)m_dwNumMat; i++) 
			{
				SAFE_RELEASE(m_ppTextures[i]);
			}
		}
		SAFE_DELETE_ARRAY(m_pMaterials);
		SAFE_DELETE_ARRAY(m_ppTextures);	
		m_lpDevice = NULL;
	}
}

HRESULT		XMesh::AddRef()
{
	return m_dwRef++;
}

HRESULT		XMesh::Release()
{
	DWORD dwID = m_dwID;
	if(m_dwRef > 0)
	{
		m_dwRef--;		
	}
	if(m_dwRef == 0)
	{			
		delete this;
		TRACE("XMesh object destroyed. ID=%d", dwID);
		return 0;
	}	
	else
	{
		return m_dwRef;
	}	
}

int		XMesh::ReleaseResource()
{
	SAFE_RELEASE(m_lpMesh);
	return S_OK;
}

int		XMesh::LoadMeshFromFile(LPCSTR szFileName, DWORD dwFlags)
{
	LPD3DXBUFFER pD3DXMtrlBuffer;
	LPD3DXMESH lpMesh;
	DWORD dwRet;
	IDisplay *dis = NULL;	
	CreateDisplayObject(DISIID_IDisplay, (void**) &dis);
	if(!dis)
	{
		XException exp("LoadMeshFromFile: Device is invalid.");
		THROW(exp);			
	}
	// Get device from IDisplay
	m_lpDevice = dis->GetDevice();
		

	if(!m_lpMesh) {

		// create full path to the texture
		if(!cmnGetDataFilePath(m_szFilePath, szFileName)) 
		{
			TRACE("Cannot create full path to the texture.");
		}
		// load mesh
		if(D3D_OK == (dwRet = D3DXLoadMeshFromX(m_szFilePath, 0, m_lpDevice, NULL, &pD3DXMtrlBuffer, &m_dwNumMat, &m_lpMesh))) 
		{
			TRACE("Mesh '%s' was loaded.", szFileName);
			// get material buffer
			D3DXMATERIAL* d3dxMaterials = (D3DXMATERIAL*)pD3DXMtrlBuffer->GetBufferPointer();
			// create arrays of materials and textures (only pointers)
			m_ppTextures = new ITexture*[m_dwNumMat];
			m_pMaterials = new D3DMATERIAL8[m_dwNumMat];
			//
			// get materials and load textures
			for(int i = 0; i < (int)m_dwNumMat; i++) 
			{
				m_pMaterials[i] = d3dxMaterials[i].MatD3D;
				m_pMaterials[i].Ambient = m_pMaterials[i].Diffuse;

				if(d3dxMaterials[i].pTextureFilename)
				{					
					if(S_OK == CreateDisplayObject(DISIID_ITexture, (void**)&m_ppTextures[i]))
					{
						m_ppTextures[i]->LoadTextureFromFile(d3dxMaterials[i].pTextureFilename);					
					}
				}			
				else
				{
					m_ppTextures[i] = NULL;
				}
			}
			pD3DXMtrlBuffer->Release();
			//
			// Add normals
			dwRet = m_lpMesh->CloneMeshFVF(m_lpMesh->GetOptions(), MESH_FORMAT, m_lpDevice, &lpMesh);
			if(dwRet != D3D_OK)
			{
				XException exp("Cannot clone mesh.", dwRet);
				THROW(exp);
			}
			if(!(m_lpMesh->GetFVF() & D3DFVF_NORMAL))
			{
				// Compute normals
				dwRet = D3DXComputeNormals(lpMesh, NULL);
				if(dwRet != D3D_OK)
				{
					XException exp("Cannot compute normals.", dwRet);
					THROW(exp);
				}
			}
			SAFE_RELEASE(m_lpMesh);
			//
			// Generate atr table
			dwRet = lpMesh->Optimize(D3DXMESHOPT_ATTRSORT, NULL, NULL, NULL, NULL, &m_lpMesh);
			if(dwRet != S_OK)
			{
				XException exp("Cannot optimize mesh!", dwRet);
				THROW(exp);
			}
			SAFE_RELEASE(lpMesh);			
		}
		else
		{			
			XException exp("Mesh wasn't loaded!", dwRet);
			THROW(exp);			
		}
		// save par to restore
		m_meshType = CUSTOM;	
		// Add this texture to the manager
		dis->GetResourceManager()->AddResource(this);
		SAFE_RELEASE(dis)
		return 0;
	}
	SAFE_RELEASE(dis);
	return -1;
}


int		XMesh::SetSubsetColor(int iSubSet, D3DCOLOR dwColor)
{
	MESH_VERTEX *pVertices;
	DWORD dwSize, dwRet;
	int i, v;
	D3DXATTRIBUTERANGE *pAtr;
	if(m_lpDevice && int(m_dwNumMat) > iSubSet && m_meshType == CUSTOM)
	{
		dwRet = m_lpMesh->GetAttributeTable(NULL, &dwSize);
		if(dwRet != S_OK)
		{
			XException exp("Cannot get atr table size!", dwRet);
			THROW(exp);
		}
		pAtr = new D3DXATTRIBUTERANGE[dwSize];
		dwRet = m_lpMesh->GetAttributeTable(pAtr, &dwSize);
		if(dwRet != S_OK)
		{
			XException exp("Cannot get atr table!", dwRet);
			THROW(exp);
		}
		if(int(dwSize) > iSubSet)
		{			
			dwRet = m_lpMesh->LockVertexBuffer(0, (BYTE**) &pVertices);
			if(dwRet != S_OK)
			{
				XException exp("Cannot lock mesh VB!", dwRet);
				THROW(exp);
			}
			// Find first vertex for this subset
			v = pAtr[iSubSet].VertexStart;
			for(i = 0; i < int(pAtr[iSubSet].VertexCount); i++)
			{
				// Modify color
				pVertices[v].dwDiffuse = dwColor;
				v++;
			}

			dwRet = m_lpMesh->UnlockVertexBuffer();
			if(dwRet != S_OK)
			{
				XException exp("Cannot unlock mesh VB!", dwRet);
				THROW(exp);
			}
		}
		SAFE_DELETE_ARRAY(pAtr);		
	}	
	return -1;
}

int		XMesh::CreateSphere(float fRadius, UINT uSlices, UINT uStacks)
{
	IDisplay *dis = NULL;
	CreateDisplayObject(DISIID_IDisplay, (void**) &dis);
	if(!dis)
	{
		XException exp("LoadTextureFromFile: Device is invalid.");
		THROW(exp);			
	}
	// Get device from IDisplay
	m_lpDevice = dis->GetDevice();
	
	if(!m_lpMesh) {

		if(D3D_OK == D3DXCreateSphere(m_lpDevice, fRadius, uSlices, uStacks, &m_lpMesh, NULL)) 
		{
			TRACE("Sphere was loaded.");
		}
		else
		{
			TRACE("Sphere wasn't loaded.");
		}
		// save par to restore
		m_meshType = SPHERE;		
		m_fPar1 = fRadius;
		m_uPar4 = uSlices;
		m_uPar5 = uStacks;
		// Add this texture to the manager
		dis->GetResourceManager()->AddResource(this);
		SAFE_RELEASE(dis)
		return 0;
	}
	SAFE_RELEASE(dis);
	return -1;
}

int		XMesh::CreateBox(float fWidth, float fHeight, float fDepth)
{
	IDisplay *dis = NULL;
	CreateDisplayObject(DISIID_IDisplay, (void**) &dis);
	if(!dis)
	{
		XException exp("LoadTextureFromFile: Device is invalid.");
		THROW(exp);			
	}
	// Get device from IDisplay
	m_lpDevice = dis->GetDevice();
	
	if(!m_lpMesh) {

		if(D3D_OK == D3DXCreateBox(m_lpDevice, fWidth, fHeight, fDepth, &m_lpMesh, NULL)) 
		{
			TRACE("Box was loaded.");
		}
		else
		{
			TRACE("Box wasn't loaded.");
		}
		// save par to restore
		m_meshType = BOX;		
		m_fPar1 = fWidth;
		m_fPar2 = fHeight;
		m_fPar3 = fDepth;
		// Add this texture to the manager
		dis->GetResourceManager()->AddResource(this);
		SAFE_RELEASE(dis)
		return 0;
	}
	SAFE_RELEASE(dis);
	return -1;
}

int		XMesh::CreateCylinder(float fRadius1, float fRadius2, float fLenght, UINT uSlices, UINT uStacks)
{
	IDisplay *dis = NULL;
	CreateDisplayObject(DISIID_IDisplay, (void**) &dis);
	if(!dis)
	{
		XException exp("LoadTextureFromFile: Device is invalid.");
		THROW(exp);			
	}
	// Get device from IDisplay
	m_lpDevice = dis->GetDevice();
	
	if(!m_lpMesh) {

		if(D3D_OK == D3DXCreateCylinder(m_lpDevice, fRadius1, fRadius2, fLenght, uSlices, uStacks, &m_lpMesh, NULL)) 
		{
			TRACE("Cylinder was loaded.");
		}
		else
		{
			TRACE("Cylinder wasn't loaded.");
		}
		// save par to restore
		m_meshType = CYLINDER;		
		m_fPar1 = fRadius1;
		m_fPar2 = fRadius2;
		m_fPar3 = fLenght;
		m_uPar4 = uSlices;
		m_uPar5 = uStacks;
		// Add this texture to the manager
		dis->GetResourceManager()->AddResource(this);
		SAFE_RELEASE(dis)
		return 0;
	}
	SAFE_RELEASE(dis);
	return -1;
}

int		XMesh::CreateTeapot()
{
	IDisplay *dis = NULL;
	CreateDisplayObject(DISIID_IDisplay, (void**) &dis);
	if(!dis)
	{
		XException exp("LoadTextureFromFile: Device is invalid.");
		THROW(exp);			
	}
	// Get device from IDisplay
	m_lpDevice = dis->GetDevice();
	
	if(!m_lpMesh) {

		if(D3D_OK == D3DXCreateTeapot(m_lpDevice, &m_lpMesh, NULL)) 
		{
			TRACE("Teapot was loaded.");
		}
		else
		{
			TRACE("Teapot wasn't loaded.");
		}
		// save par to restore
		m_meshType = TEAPOT;	
		// Add this texture to the manager
		dis->GetResourceManager()->AddResource(this);
		SAFE_RELEASE(dis)
		return 0;
	}
	SAFE_RELEASE(dis);
	return -1;
}

int		XMesh::CreateTorus(float fInnerRadius, float fOuterRadius, UINT uSides, UINT uRings)
{
	IDisplay *dis = NULL;
	CreateDisplayObject(DISIID_IDisplay, (void**) &dis);
	if(!dis)
	{
		XException exp("LoadTextureFromFile: Device is invalid.");
		THROW(exp);			
	}
	// Get device from IDisplay
	m_lpDevice = dis->GetDevice();
	
	if(!m_lpMesh) {

		if(D3D_OK == D3DXCreateTorus(m_lpDevice, fInnerRadius, fOuterRadius, uSides, uRings, &m_lpMesh, NULL)) 
		{
			TRACE("Torus was loaded.");
		}
		else
		{
			TRACE("Torus wasn't loaded.");
		}
		// save par to restore
		m_meshType = TORUS;		
		m_fPar1 = fInnerRadius;
		m_fPar2 = fOuterRadius;
		m_uPar4 = uSides;
		m_uPar5 = uRings;
		// Add this texture to the manager
		dis->GetResourceManager()->AddResource(this);
		SAFE_RELEASE(dis)
		return 0;
	}
	SAFE_RELEASE(dis);
	return -1;
}

int	XMesh::Draw()
{
	if(m_lpDevice && m_bVisible && m_lpMesh)
	{
		// Set transformation matrix before rendering
		if(&m_TransMat)
		{
			m_lpDevice->SetTransform(D3DTS_WORLD, &m_TransMat);
		}
		if(m_meshType == CUSTOM) 
		{
			for(int i = 0; i < (int)m_dwNumMat; i++) 
			{
				m_lpDevice->SetMaterial(&m_pMaterials[i]);
				m_lpDevice->SetTexture(0, (m_ppTextures[i]) ? m_ppTextures[i]->GetTexture() : NULL);
				m_lpMesh->DrawSubset(i);
			}
		}
		else 
		{			
			// set default material
			D3DMATERIAL8 mtrl;
			ZeroMemory( &mtrl, sizeof(D3DMATERIAL8) );
			mtrl.Diffuse.r = mtrl.Ambient.r = 1.0f;
			mtrl.Diffuse.g = mtrl.Ambient.g = 1.0f;
			mtrl.Diffuse.b = mtrl.Ambient.b = 1.0f;
			mtrl.Diffuse.a = mtrl.Ambient.a = 1.0f;
			m_lpDevice->SetMaterial(&mtrl);

			m_lpDevice->SetTexture(0, NULL);
			m_lpMesh->DrawSubset(0);
		}
		// set indentity matrix
		if(&m_TransMat)
		{
			D3DXMATRIX mat;
			D3DXMatrixIdentity(&mat);
			m_lpDevice->SetTransform(D3DTS_WORLD, &mat);
		}
		return 0;
	}
	return -1;
}

int		XMesh::Restore()
{
	if(m_lpDevice) 
	{
		switch(m_meshType) {
			case CUSTOM:				
				// load mesh
				if(D3D_OK == D3DXLoadMeshFromX(m_szFilePath, 0, m_lpDevice, NULL, NULL, NULL, &m_lpMesh)) 
				{
					TRACE("Mesh '%s' was loaded.", m_szFilePath);								
					return 0;
				}
				else
				{
					TRACE("Mesh '%s' wasn't loaded.", m_szFilePath);
				}
				break;
			case SPHERE:
				return CreateSphere(m_fPar1, m_uPar4, m_uPar5);
			case BOX:
				return CreateBox(m_fPar1, m_fPar2, m_fPar3);
			case CYLINDER:
				return CreateCylinder(m_fPar1, m_fPar2, m_fPar3, m_uPar4, m_uPar5);
			case TORUS:
				return CreateTorus(m_fPar1, m_fPar2, m_uPar4, m_uPar5);
			case TEAPOT:
				return CreateTeapot();
			}
	}
	return -1;
}


void	XMesh::Translate(float fX, float fY, float fZ)
{
	D3DXMATRIX	matTemp;
	// Create translation matrix
	D3DXMatrixTranslation(&matTemp, fX, fY, fZ);
	// modify transformation matrix of the object
	D3DXMatrixMultiply(&m_TransMat, &m_TransMat, &matTemp);
}

void	XMesh::RotateX(float fAngle)
{
	D3DXMATRIX	matTemp;
	// create rotation matrix
	D3DXMatrixRotationX(&matTemp, fAngle);
	// modify transformation matrix of the object
	D3DXMatrixMultiply(&m_TransMat, &m_TransMat, &matTemp);

}

void	XMesh::RotateY(float fAngle)
{
	D3DXMATRIX	matTemp;
	// create rotation matrix
	D3DXMatrixRotationY(&matTemp, fAngle);
	// modify transformation matrix of the object
	D3DXMatrixMultiply(&m_TransMat, &m_TransMat, &matTemp);
}

void	XMesh::RotateZ(float fAngle)
{
	D3DXMATRIX	matTemp;
	// create rotation matrix
	D3DXMatrixRotationZ(&matTemp, fAngle);
	// modify transformation matrix of the object
	D3DXMatrixMultiply(&m_TransMat, &m_TransMat, &matTemp);
}

void					XMesh::ScaleXYZ(float fFactor)
{
	D3DXMATRIX	matTemp;
	// Create scaling matrix
	D3DXMatrixScaling(&matTemp, fFactor, fFactor, fFactor);
	// modify transformation matrix of the object
	D3DXMatrixMultiply(&m_TransMat, &m_TransMat, &matTemp);
}

void	XMesh::ScaleX(float fFactor)
{
	D3DXMATRIX	matTemp;
	// Create scaling matrix
	D3DXMatrixScaling(&matTemp, fFactor, 1.0f, 1.0f);
	// modify transformation matrix of the object
	D3DXMatrixMultiply(&m_TransMat, &m_TransMat, &matTemp);
}

void	XMesh::ScaleY(float fFactor)
{
	D3DXMATRIX	matTemp;
	// Create scaling matrix
	D3DXMatrixScaling(&matTemp, 1.0f, fFactor, 1.0f);
	// modify transformation matrix of the object
	D3DXMatrixMultiply(&m_TransMat, &m_TransMat, &matTemp);
}

void	XMesh::ScaleZ(float fFactor)
{
	D3DXMATRIX	matTemp;
	// Create scaling matrix
	D3DXMatrixScaling(&matTemp, 1.0f, 1.0f, fFactor);
	// modify transformation matrix of the object
	D3DXMatrixMultiply(&m_TransMat, &m_TransMat, &matTemp);
}


BOOL	XMesh::operator==(IResource &pRes)
{

	return FALSE;
}